#include <string>
#include <map>
#include <math.h>
#include <stdio.h>
#include <list>
#include <iostream>
#include <getopt.h>

#include "ContigLink_AMOS.hh"
#include "ContigEdge_AMOS.hh"
#include "datatypes_AMOS.hh"
#include "Bank_AMOS.hh"
#include "BankStream_AMOS.hh"

using namespace AMOS;
using namespace std;

typedef multimap<pair<ID_t, ID_t> , ContigLink_t > LinkMap_t;

map<string, string> globals; // global variables
ID_t EdgeId = 0;             // where numbering starts for edges

void printHelpText()
{
  cerr << 
    "\n"
    "USAGE:\n"
    "\n"
    "Bundler -b[ank] <bank_name>\n"
       << endl;
}


bool GetOptions(int argc, char ** argv)
{  
  int option_index = 0;
  static struct option long_options[] = {
    {"help",  0, 0, 'h'},
    {"h",     0, 0, 'h'},
    {"b",     1, 0, 'b'},
    {"bank",  1, 0, 'b'},
    {0, 0, 0, 0}
  };
  
  int c;
  while ((c = getopt_long_only(argc, argv, "", long_options, &option_index))!= -
	 1){
    switch (c){
    case 'h':
      printHelpText();
      break;
    case 'b':
      globals["bank"] = string(optarg);
      break;
    case '?':
      return false;
    }
  }

  return true;
} // GetOptions


int main(int argc, char *argv[])
{
  LinkMap_t linkMap;

  if (!GetOptions(argc, argv)){
    cerr << "Command line parsing failed" << endl;
    printHelpText();
    exit(1);
  }

  if (globals.find("bank") == globals.end()){ // no bank was specified
    cerr << "A bank must be specified" << endl;
    exit(1);
  }


  BankStream_t link_bank (ContigLink_t::NCODE);
  try {
    if (! link_bank.exists(globals["bank"])) {
      cerr << "Cannot find a contig link account in bank: " << globals["bank"] << endl;
      exit(1);
    }
    link_bank.open(globals["bank"], B_READ);
  } catch (Exception_t & e)
    {
      cerr << "Failed to open link account in bank " << globals["bank"] 
	   << ": " << endl << e << endl;
      exit(1);
    }

  BankStream_t edge_bank (ContigEdge_t::NCODE);
  try {
    if (! edge_bank.exists(globals["bank"])) 
      edge_bank.create(globals["bank"]);
    else 
      edge_bank.open(globals["bank"]);
  } catch (Exception_t & e)
    {
      cerr << "Failed to open edge account in bank " << globals["bank"] 
	   << ": " << endl << e << endl;
      exit(1);
    }
  

  Size_t ti;
  Size_t minlen, maxlen;  // contig adjacency gap range;

  ContigLink_t ctl;

  while (link_bank >> ctl){
    ID_t ctgA, ctgB;
    
    ctgA = ctl.getContigs().first;
    ctgB = ctl.getContigs().second;

    if (linkMap.find(pair<ID_t, ID_t>(ctgB, ctgA)) != linkMap.end()){
      // must reverse the link
      ctl.flip();
    } 
    linkMap.insert(pair<pair<ID_t, ID_t>, ContigLink_t>
		   (ctl.getContigs(), ctl));
  }

  // now all the links should be in the linkMap, nicely grouped by the contigs
  // they connect.

  LinkMap_t::iterator li = linkMap.begin();
  while (li != linkMap.end()){
    pair<ID_t, ID_t> lastPair = (*li).first;

    cerr << "Doing contig pair (" << lastPair.first << ", " 
	 << lastPair.second << ")" << endl; 

    map <LinkAdjacency_t, list<LinkMap_t::iterator> > adjList;
    map <LinkAdjacency_t, list<LinkMap_t::iterator> > bestList;
    map <LinkAdjacency_t, int> adjacencies;

    while (li->first == lastPair){
      LinkAdjacency_t a = li->second.getAdjacency();

      adjacencies[a]++;
      adjList[a].push_back(li);

      li++;
    }

    cerr << "NORMALS: " << adjacencies[(LinkAdjacency_t)ContigLink_t::NORMAL]
	 << " ANTINORMALS: " << adjacencies[(LinkAdjacency_t)ContigLink_t::ANTINORMAL]
	 << " INNIES: " << adjacencies[(LinkAdjacency_t)ContigLink_t::INNIE]
	 << " OUTIES: " << adjacencies[(LinkAdjacency_t)ContigLink_t::OUTIE]
	 << "\n";

    LinkAdjacency_t bestAdj;
    int bestAdjCount = 0;

    for (map<LinkAdjacency_t, int>::iterator at = adjacencies.begin();
	 at != adjacencies.end(); at++){

      if (at->second > 1) {

	map<Size_t, LinkMap_t::iterator> begins, ends;
	list<LinkMap_t::iterator> best;
	list<LinkMap_t::iterator>::iterator bestEnd = best.end();
	Size_t bestCoord;
	int ncurrent = 0, nbest = 0;

	for (list<LinkMap_t::iterator>::iterator 
	       lnks = adjList[at->first].begin();
	     lnks != adjList[at->first].end(); lnks++) {

 	  begins.insert(pair<Size_t, LinkMap_t::iterator> 
			(((**lnks).second.getSize() - 3 * (**lnks).second.getSD()), *lnks));
	  ends.insert(pair<Size_t, LinkMap_t::iterator>
		      (((**lnks).second.getSize() + 3 * (**lnks).second.getSD()), *lnks));
	}

	map<Size_t, LinkMap_t::iterator>::iterator bgi = begins.begin();
	map<Size_t, LinkMap_t::iterator>::iterator eni = ends.begin();
	
	while (bgi != begins.end() || eni != ends.end()){
	  if (bgi != begins.end() && bgi->first <= eni->first) {

	    Size_t bgiLend = bgi->second->second.getSize() 
	      - 3 * bgi->second->second.getSD(),
	      bgiRend = bgi->second->second.getSize() 
	      + 3 * bgi->second->second.getSD();

	    ncurrent++;

	    if (ncurrent > nbest) {

	      nbest = ncurrent;
	      bestEnd = best.begin();
	      bestCoord = bgiLend;
	    }

	    bgi++;
	  } 
	  else if (eni != ends.end() && 
		   (bgi == begins.end() || 
		    bgi->first > eni->first)) {

	    Size_t eniLend = eni->second->second.getSize() 
	      - 3 * eni->second->second.getSD(),
	      eniRend = eni->second->second.getSize() 
	      + 3 * eni->second->second.getSD();

	    if (eniLend <= bestCoord && eniRend >= bestCoord){ 

	      best.push_front(eni->second);
	    } 

	    ncurrent--;
	    eni++;
	  } 
	} 

	at->second = nbest;
	Size_t minRange, maxRange;
	list<LinkMap_t::iterator>::iterator bg = best.begin();
	Size_t sz = (**bg).second.getSize();
	Size_t sd = (**bg).second.getSD();
	minRange = sz - 3 * sd;
	maxRange = sz + 3 * sd;

	for (list<LinkMap_t::iterator>::iterator 
	       bi = adjList[at->first].begin();
	     bi != adjList[at->first].end(); bi++){
	  bestList[at->first].push_back(*bi);
	}

	for (list<LinkMap_t::iterator>::iterator bi = best.begin();
	     bi != bestEnd; bi++) {

	  //bestList[at->first].push_back(*bi);
	  sz = (**bi).second.getSize();
	  sd = (**bi).second.getSD();

	  if (sz - 3 * sd > minRange)
	    minRange = sz - 3 * sd;
	  if (sz + 3 * sd < maxRange)
	    maxRange = sz + 3 * sd;
	}
      } 
      else {
	
	for (list<LinkMap_t::iterator>::iterator 
	       bi = adjList[at->first].begin();
	     bi != adjList[at->first].end(); bi++){
	  bestList[at->first].push_back(*bi);
	}
      }

      if(bestList[at->first].size() == 0)
	continue;

      ContigEdge_t cte;

      bestAdjCount = at->second;
      bestAdj = at->first;

      cte.setContigs(lastPair);
      cte.setAdjacency(bestAdj);
      
      vector<ID_t> linkIds;
      linkIds.reserve(bestList[bestAdj].size());

      float newMean, newSD, p = 0, q = 0;
      for (list<LinkMap_t::iterator>::iterator ni = bestList[bestAdj].begin();
	   ni != bestList[bestAdj].end(); ni++){
	float tmp = (**ni).second.getSD();
	tmp *= tmp;
	
	p += (**ni).second.getSize() / tmp;
	q += 1 / tmp; 
	linkIds.push_back((**ni).second.getIID());
      }
      
      newMean = p / q;
      newSD = 1 / sqrt(q);
      
      cte.setSize((int)newMean);
      cte.setSD((int)newSD);

      cte.setContigLinks(linkIds);
      cte.setIID(++EdgeId);
      
      edge_bank << cte;
    }
  }

  edge_bank.close();
  link_bank.close();
  return(0);
}

