#include "foundation_AMOS.hh"
#include <cstdlib>
#include <map>
#include <string>
#include <iostream>
#include <unistd.h>

using namespace std;
using namespace AMOS;

char* OPT_BankName;
int OPT_LinkThreshold;
int OPT_ContigLengthThreshold;

void PrintUsage(const char* s) {

  cerr
    << "\nUSAGE: " << s << "  [-b bank-path] [-l link-threshold] [-c contig-length-threshold]\n\n";
  return;
}

void PrintHelp(const char* s) {

  PrintUsage (s);
  cerr
    << "-b path       Set path of bank\n"
    << "-l value      Threshold value for minimum number of links\n"
    << "-c value      Threshold value for minimum contig length\n"
    << endl;
  
  cerr
    << "Filters unwanted contig edges from a bank\n"
    << endl;

  return;
}

void ParseArgs(int argc, char** argv) {

  int ch, errflg = 0;
  optarg = NULL;
  
  while(!errflg  &&
	((ch = getopt(argc, argv, "b:l:c:h")) != EOF))
    switch (ch) {

    case 'b':
      OPT_BankName = optarg;
      break;

    case 'l':
      OPT_LinkThreshold = atoi(optarg);
      break;

    case 'c':
      OPT_ContigLengthThreshold = atoi(optarg);
      break;
      
    case 'h':
      PrintHelp(argv[0]);
	exit(EXIT_SUCCESS);
	break;
	
    default:
      errflg ++;
    }

  if(errflg > 0 || optind != argc) {

    PrintUsage(argv[0]);
    cerr << "Try '" << argv[0] << " -h' for more information.\n";
    exit(EXIT_FAILURE);
  }
}

int main(int argc, char ** argv) {

  ParseArgs (argc, argv);

  BankStream_t contigs(Contig_t::NCODE);
  contigs.open(OPT_BankName);

  map<ID_t, bool> ignore_contig;
  Contig_t contig;
  while(contigs >> contig) {

    ignore_contig[contig.getIID()] = true;

    if(contig.getReadTiling().size() < 5)
      cerr << contig.getIID() << " has too few reads: " << contig.getReadTiling().size() << endl;
    else if(contig.getUngappedLength() < OPT_ContigLengthThreshold)
      cerr << contig.getIID() << " is too small: " << contig.getUngappedLength() << endl;
    else
      ignore_contig[contig.getIID()] = false;
  }
  contigs.close();

  BankStream_t edges(ContigEdge_t::NCODE);
  ContigEdge_t edge; double omitted = 0;
  map<ID_t, int> forward_adj; map<ID_t, int> back_adj;
 
  edges.open(OPT_BankName);
  while(edges >> edge) {

    pair<ID_t, ID_t> contigs = edge.getContigs();

    if(ignore_contig[contigs.first] || ignore_contig[contigs.second] ||
       edge.getContigLinks().size() < OPT_LinkThreshold ||
       edge.getSize() == 0 && edge.getContigLinks().size() < 5) {

      omitted++;
      edges.remove(edge.getIID());
      continue;
    }
    else if((edge.getContigLinks().size() == 1) && 
	    (int(edge.getSize())+4*int(edge.getSD()) < -500)) {

      omitted++;
      cerr << "The size estimate from this edge suggests a substantial overlap: " << contigs.first << " " << contigs.second << " " << edge.getSize() << " "
	   << edge.getSD() << " " << edge.getContigLinks().size() << " " << endl;
      edges.remove(edge.getIID());
      continue;
    }

    if(edge.getAdjacency() == ContigLink_t::NORMAL ||
       edge.getAdjacency() == ContigLink_t::INNIE)
      forward_adj[contigs.first]++;
    else
      back_adj[contigs.first]++;
      
    edge.flip();

    if(edge.getAdjacency() == ContigLink_t::NORMAL ||
       edge.getAdjacency() == ContigLink_t::INNIE)
      forward_adj[contigs.second]++;
    else
      back_adj[contigs.second]++;
  }

  cerr << "Number of edges removed: " << omitted << endl;
  edges.clean();
  edges.close();

  edges.open(OPT_BankName);
  while(edges >> edge) {

    pair<ID_t, ID_t> contigs = edge.getContigs();

    if(edge.getAdjacency() == ContigLink_t::NORMAL ||
       edge.getAdjacency() == ContigLink_t::INNIE) { 
      if(forward_adj[contigs.first] > 1)
	edges.remove(edge.getIID());
    }
    else if(back_adj[contigs.first] > 1)
      edges.remove(edge.getIID());

    edge.flip();

    if(edges.existsIID(edge.getIID()))
      if(edge.getAdjacency() == ContigLink_t::NORMAL ||
	 edge.getAdjacency() == ContigLink_t::INNIE) { 
	if(forward_adj[contigs.second] > 1)
	  edges.remove(edge.getIID());
      }
      else if(back_adj[contigs.second] > 1)
	edges.remove(edge.getIID());
  }

  edges.clean();
  edges.close();

  return EXIT_SUCCESS;
}
