#include <vector>
#include <unistd.h>
#include "itoa.h"
#include "foundation_AMOS.hh"

using namespace std;
using namespace AMOS;

string OPT_BankName;

void PrintUsage(const char* s) {

  cerr
    << "\nUSAGE: " << s << "  [-b bank] \n\n";
  return;
}

void PrintHelp(const char* s) {

  PrintUsage (s);
  cerr
    << "-b path       Set path of contigs to scaffold\n"
    << "-h            Display help information\n"
    << endl;

  cerr
    << "Scaffold contigs based on edge information\n"
    << endl;

  return;
}

void ParseArgs(int argc, char** argv) {

  int ch, errflg = 0;
  optarg = NULL;
  
  while(!errflg  &&
	((ch = getopt(argc, argv, "b:h")) != EOF))
    switch (ch) {

    case 'b':
      OPT_BankName = optarg;
      break;
      
    case 'h':
      PrintHelp(argv[0]);
	exit(EXIT_SUCCESS);
	break;
	
    default:
      errflg ++;
    }

  if(errflg > 0 || optind != argc) {

    PrintUsage(argv[0]);
    cerr << "Try '" << argv[0] << " -h' for more information.\n";
    exit(EXIT_FAILURE);
  }
}

Tile_t make_tile(ID_t id, bool forward, Pos_t offset, Size_t length) {

  Tile_t tile; tile.source = id; 
  tile.offset = offset;
  tile.range = (forward ? Range_t(0, length) : Range_t(length, 0));

  return tile;
}

int main(int argc, char ** argv) {

  ParseArgs (argc, argv);

  BankStream_t contigs(Contig_t::NCODE);
  contigs.open(OPT_BankName);
  Contig_t contig;

  map<ID_t, Size_t> contig_length;
  while(contigs >> contig) 
    contig_length[contig.getIID()] = contig.getLength();
  contigs.close();

  BankStream_t edges(ContigEdge_t::NCODE);
  edges.open(OPT_BankName);
  ContigEdge_t edge;

  map<ID_t, map<ID_t, ContigEdge_t> > adj;
  map<ID_t, int> forward_adj; map<ID_t, int> back_adj;
  map<ID_t, bool> used;
  while(edges >> edge) {

    pair<ID_t, ID_t> contigs = edge.getContigs();
    adj[contigs.first][contigs.second] = edge;

    if(edge.getAdjacency() == ContigLink_t::NORMAL ||
       edge.getAdjacency() == ContigLink_t::INNIE)
      forward_adj[contigs.first]++;
    else
      back_adj[contigs.first]++;
      
    edge.flip(); adj[contigs.second][contigs.first] = edge;

    if(edge.getAdjacency() == ContigLink_t::NORMAL ||
       edge.getAdjacency() == ContigLink_t::INNIE)
      forward_adj[contigs.second]++;
    else
      back_adj[contigs.second]++;

    used[contigs.first] = false; used[contigs.second] = false;
  }
  edges.close();

  BankStream_t scaffolds(Scaffold_t::NCODE);
  scaffolds.create(OPT_BankName);

  int scaff_id = 0;

  for(map<ID_t, map<ID_t, ContigEdge_t> >::iterator it = adj.begin(); it != adj.end(); it++) {

    if(used[it->first] || it->second.size() != 1)
      continue;

    Scaffold_t scaff; Pos_t offset = 0;    
    scaff_id++; scaff.setIID(scaff_id); scaff.setEID(string("scaff")+string(itoa(scaff_id))); 
    ID_t curr_contig = it->first;
    bool forward = ((adj[curr_contig].begin())->second.getAdjacency() == ContigLink_t::NORMAL) ||
      ((adj[curr_contig].begin())->second.getAdjacency() == ContigLink_t::INNIE);

    while(1) {

      used[curr_contig] = true;
      scaff.getContigTiling().push_back(make_tile(curr_contig, forward, offset, contig_length[curr_contig]));
      offset += contig_length[curr_contig];

      if(adj[curr_contig].size() == 0)
	break;

      if(forward_adj[(adj[curr_contig].begin())->first] <= 1 &&
	 back_adj[(adj[curr_contig].begin())->first] <= 1) {
	
	scaff.getContigEdges().push_back((adj[curr_contig].begin())->second.getIID());
	adj[(adj[curr_contig].begin())->first].erase(curr_contig);
	offset += ((adj[curr_contig].begin())->second.getSize() > 0 ? (adj[curr_contig].begin())->second.getSize() : 0);
	
	forward = ((adj[curr_contig].begin())->second.getAdjacency() == ContigLink_t::NORMAL) ||
	  ((adj[curr_contig].begin())->second.getAdjacency() == ContigLink_t::OUTIE);
      }
      else
	break;
      
      curr_contig = (adj[curr_contig].begin())->first;
    } 

    scaffolds << scaff;
  }
    
  for(map<ID_t, Size_t>::iterator it = contig_length.begin(); it != contig_length.end(); it++)
    if(!used[it->first]) {
      
      Scaffold_t scaff;          
      scaff_id++; scaff.setIID(scaff_id); scaff.setEID(string("scaff")+string(itoa(scaff_id))); 
      scaff.getContigTiling().push_back(make_tile(it->first, true, 0, it->second));
      scaffolds << scaff;
    }

  scaffolds.close();

  return EXIT_SUCCESS;
}
