#include "foundation_AMOS.hh"
#include "delta.hh"
#include <cstdlib>
#include <map>
#include <string>
#include <iostream>
#include <unistd.h>

using namespace std;
using namespace AMOS;

string OPT_ContigBankName;
string OPT_ReadBankName;
string OPT_DeltaFile;
string OPT_ResultBankName;

void PrintUsage(const char* s) {

  cerr
    << "\nUSAGE: " << s << "  [-c contig-bank] [-r read-bank] [-d delta-file] [-o output-bank]\n\n";
  return;
}

void PrintHelp(const char* s) {

  PrintUsage (s);
  cerr
    << "-c path       Set path of the AMOS bank of reference contigs to use\n"
    << "-r path       Set path of the AMOS bank of reads to add\n"
    << "-d path       Set path of delta file to use\n"
    << "-o path       Set path of the AMOS bank of resulting contigs\n"
    << "-h            Display help information\n"
    << endl;

  cerr
    << "Add reads to contigs based on the alignment\n"
    << "information contained in the delta file.\n" 
    << endl;

  return;
}

void ParseArgs(int argc, char** argv) {

  int ch, errflg = 0;
  optarg = NULL;
  
  while(!errflg  &&
	((ch = getopt(argc, argv, "c:r:d:o:h")) != EOF))
    switch (ch) {

    case 'c':
      OPT_ContigBankName = optarg;
      break;
      
    case 'r':
      OPT_ReadBankName = optarg;
      break;
      
    case 'd':
      OPT_DeltaFile = optarg;
      break;
      
    case 'o':
      OPT_ResultBankName = optarg;
      break;

    case 'h':
      PrintHelp(argv[0]);
	exit(EXIT_SUCCESS);
	break;
	
    default:
      errflg ++;
    }

  if(errflg > 0 || optind != argc) {

    PrintUsage(argv[0]);
    cerr << "Try '" << argv[0] << " -h' for more information.\n";
    exit(EXIT_FAILURE);
  }
}

const Tile_t& makeTile(ID_t id, DeltaAlignment_t alignment) { 

      static Tile_t tile; tile.source = id; 
      if(alignment.eR < alignment.sR)
	cerr << "TILE: " << alignment.sR << " " << alignment.eR << endl;

      tile.offset = alignment.sR-1;
      tile.range = Range_t(alignment.sQ, alignment.eQ);
      tile.gaps.clear();

      return tile;
}

inline int max(int a, int b) { return (a > b ? a : b); }
inline int min(int a, int b) { return (a < b ? a : b); }

const double score_threshold = 0.90;

bool end_match(DeltaRecord_t& record) {

  DeltaAlignment_t& align = record.aligns[0];

  if(align.eQ >= align.sQ)
    return ((int(align.sR-align.sQ) < 0) || 
	    (int(record.lenR) < int(align.eR+(record.lenQ-align.eQ))));
  else
    return ((int(align.sR-(record.lenQ-align.sQ)) < 0) ||
	    (int(record.lenR) < int(align.eR+align.eQ)));
}

int should_align;
int did_align;
int did_mismatch;

double score_alignment(DeltaRecord_t& record) {

  DeltaAlignment_t& align = record.aligns[0];
  Pos_t start, end; 

  if(align.eQ >= align.sQ) {

    start = max(0, align.sR-align.sQ);
    end = min(record.lenR, align.eR+(record.lenQ-align.eQ));
  }
  else {

    start = max(0, align.sR-(record.lenQ-align.sQ));
    end = min(record.lenR, align.eR+align.eQ);
  }

  should_align = end-start;
  did_align = abs((Size_t)(align.eQ-align.sQ))-align.idyc;
  did_mismatch = align.idyc;

  return did_align/double(did_align+did_mismatch);
}

#include "itoa.h"

int link_id = 0;

ContigLink_t get_join_link(ID_t read, vector<DeltaRecord_t>& d) {

  ContigLink_t link;
  link.setType(ContigLink_t::OVERLAP); link.setSource(pair<ID_t, NCode_t>(read, Read_t::NCODE));
  link_id++; link.setIID(link_id); link.setEID(itoa(link_id));

  Range_t r0(d[0].aligns[0].sQ, d[0].aligns[0].eQ);
  Range_t r1(d[1].aligns[0].sQ, d[1].aligns[0].eQ);

  if(r0.getLo() > r1.getLo()) {

    swap(r0, r1);
    swap(d[0], d[1]);
  }
  
  link.setContigs(pair<ID_t, ID_t>(atoi(d[0].idR.c_str()), atoi(d[1].idR.c_str())));
  Pos_t end = r0.getHi() + (r0.isReverse() ? d[0].aligns[0].sR : 
			    d[0].lenR-d[0].aligns[0].eR);
  Pos_t start = r1.getLo() - (r1.isReverse() ? d[1].lenR-d[1].aligns[0].eR :
			      d[1].aligns[0].sR); 
  
  link.setSize(start-end); link.setSD(SD_t(5+abs(start-end)*0.05));
  
  if(r0.isReverse() && !r1.isReverse())
    link.setAdjacency(ContigLink_t::OUTIE);
  else if(!r0.isReverse() && r1.isReverse())
    link.setAdjacency(ContigLink_t::INNIE);
  else if(!r0.isReverse() && !r1.isReverse())
    link.setAdjacency(ContigLink_t::NORMAL);
  else
    link.setAdjacency(ContigLink_t::ANTINORMAL);

  return link;
}

int main(int argc, char ** argv)
{
  try{

  ParseArgs (argc, argv);

  BankStream_t contig_read_stream(Read_t::NCODE);
  contig_read_stream.open(OPT_ContigBankName);
  BankStream_t read_out(Read_t::NCODE);
  read_out.create(OPT_ResultBankName);
  read_out.concat(contig_read_stream);

  BankStream_t read_stream(Read_t::NCODE);
  read_stream.open(OPT_ReadBankName);
  ID_t max_iid = contig_read_stream.getMaxIID();

  Read_t read;  
  while(read_stream >> read) {

    read.setIID(read.getIID()+max_iid);
    read_out << read;
  }

  read_out.close();
  read_stream.close();
  contig_read_stream.close();

  BankStream_t  contig_stream(Contig_t::NCODE);
  contig_stream.open(OPT_ContigBankName);

  Contig_t contig;
  map<ID_t, bool> ignore_contig;
  while(contig_stream >> contig) {

    ignore_contig[contig.getIID()] = true;

    if(contig.getReadTiling().size() < 5)
      cerr << contig.getIID() << " has too few reads: " << contig.getReadTiling().size() << endl;
    else if(contig.getUngappedLength() < 500)
      cerr << contig.getIID() << " is too small :" << contig.getUngappedLength() << endl;
    else
      ignore_contig[contig.getIID()] = false;
  }
  contig_stream.close();

  DeltaReader_t dr; 
  map<ID_t, vector<DeltaRecord_t> > delta_map;

  dr.open(OPT_DeltaFile);
  while(dr.readNext()) {

    DeltaRecord_t current = dr.getRecord();

    if(ignore_contig[atoi(current.idR.c_str())])
      continue;

    delta_map[atoi(current.idQ.c_str())].push_back(current);
  }
  dr.close();

  BankStream_t links(ContigLink_t::NCODE);
  links.create(OPT_ResultBankName);

  map<ID_t, vector<Tile_t> > new_tiles; int placed_count = 0;
  for(map<ID_t, vector<DeltaRecord_t> >::iterator delta_it = delta_map.begin();
      delta_it != delta_map.end(); delta_it++) {

    ID_t curr_read_id = delta_it->first;

    if(delta_it->second.size() > 2) {

      cerr << "ERROR: " << curr_read_id << " has too many matches: " 
	   << delta_it->second.size() << " (";

      for(int i = 0; i < delta_it->second.size(); i++)
	cerr << " " << delta_it->second[i].idR 
	     << "," << ignore_contig[atoi(delta_it->second[i].idR.c_str())];

      cerr << ")" << endl;

      continue;
    } 

    DeltaRecord_t best = delta_it->second[0];

    if(delta_it->second.size() > 1)
      if(end_match(delta_it->second[0]) &&
	 end_match(delta_it->second[1])) {
	
	cerr << "JOIN: " << delta_it->second[0].idR 
	     << " " << delta_it->second[0].aligns[0].sQ << " " << delta_it->second[0].aligns[0].eQ 
	     << ", " << delta_it->second[1].idR 
	     << " " << delta_it->second[1].aligns[0].sQ << " " << delta_it->second[1].aligns[0].eQ << endl;

	ContigLink_t link = get_join_link(curr_read_id, delta_it->second); 
	links << link;

	if(score_alignment(delta_it->second[0]) < score_alignment(delta_it->second[1]))
	  best = delta_it->second[1];
      }
      else {

	cerr << "ERROR: " << curr_read_id << " has too many matches: " 
	     << delta_it->second.size() << " (";
	
	cerr << " " << delta_it->second[0].idR 
	     << "," << ignore_contig[atoi(delta_it->second[0].idR.c_str())]
	     << " " << delta_it->second[0].aligns[0].sQ << " " << delta_it->second[0].aligns[0].eQ 
	     << " " << delta_it->second[0].aligns[0].sR << " " << delta_it->second[0].aligns[0].eR << " " << delta_it->second[0].lenR  
	     << ": " << delta_it->second[1].idR
	     << "," << ignore_contig[atoi(delta_it->second[1].idR.c_str())] 
	     << " " << delta_it->second[1].aligns[0].sQ << " " << delta_it->second[1].aligns[0].eQ 
	     << " " << delta_it->second[1].aligns[0].sR << " " << delta_it->second[1].aligns[0].eR << " " << delta_it->second[1].lenR
	     << ")" << endl;

	continue;
      }
    
    if(delta_it->second[0].aligns.size() > 1) {

      cerr << "Multiple within contig matches " << curr_read_id << " " 
	   << delta_it->second[0].aligns[0].sR << ","
	   << delta_it->second[0].aligns[0].eR << ","
	   << delta_it->second[0].aligns[1].sR << ","
	   << delta_it->second[0].aligns[1].eR << ","
	   << endl;
      for(int i = 1; i < delta_it->second[0].aligns.size(); i++)
	if(abs((long int)(delta_it->second[0].aligns[0].eR-delta_it->second[0].aligns[0].sR)) < 
	   abs((long int)(delta_it->second[0].aligns[i].eR-delta_it->second[0].aligns[i].sR)))
	  swap(delta_it->second[0].aligns[0], delta_it->second[0].aligns[i]);
    }

    if(score_alignment(best) > score_threshold) {     
      new_tiles[atoi(best.idR.c_str())].
	push_back(makeTile(curr_read_id+max_iid, best.aligns[0]));

      placed_count++;
    }
    else
      cerr << "ERROR: Weak alignment for " << curr_read_id 
	   << " with " << best.idR << " " << score_alignment(best) << " " 
	   << should_align << " " << did_align << " " << did_mismatch << endl;

  }

  links.close();

  BankStream_t out(Contig_t::NCODE);
  out.create(OPT_ResultBankName);

  contig_stream.open(OPT_ContigBankName);
  while(contig_stream >> contig) {

    if(!new_tiles[contig.getIID()].empty()) {

      vector<Tile_t>& current_tiling = contig.getReadTiling();
      vector<Tile_t>& to_add = new_tiles[contig.getIID()];
      for(vector<Tile_t>::iterator it = to_add.begin(); it !=  to_add.end(); it++)
      	current_tiling.push_back(*it);
    }

    out << contig;
  }
  contig_stream.close();
  out.close();

  cerr << "Placed: " << placed_count << endl;

  BankStream_t read_lib_stream(Library_t::NCODE);
  read_lib_stream.open(OPT_ReadBankName);
  BankStream_t lib_out(Library_t::NCODE);
  lib_out.create(OPT_ResultBankName);
  lib_out.concat(read_lib_stream);
  read_lib_stream.close();
  lib_out.close();

  BankStream_t  read_frag_stream(Fragment_t::NCODE);
  read_frag_stream.open(OPT_ReadBankName);
  BankStream_t frag_out(Fragment_t::NCODE);
  frag_out.create(OPT_ResultBankName);

  Fragment_t frag;
  while(read_frag_stream >> frag) {

    pair<ID_t, ID_t> mates = frag.getMatePair();
    frag.setReads(pair<ID_t, ID_t>(mates.first+max_iid, mates.second+max_iid));
    frag_out << frag;
  }

  read_frag_stream.close();
  frag_out.close();

  }
  catch(const Exception_t & e) {
   
    cerr << e.what( ) << endl;
    exit(EXIT_FAILURE);
  }

  return EXIT_SUCCESS;
}
